PARSED_GTF_PATH_GRCh38 = "Homo_sapiens.GRCh38.112.parsed.gtf"
WHOLE_GENE_LOCS_PATH_GRCh38 = "Homo_sapiens.GRCh38.112.gtf.gene_names.sorted.cleaned.bed"
REFERENCE_GENOME_GRCh38 = "ref/Homo_sapiens.GRCh38.dna.primary_assembly.fa"
REFERERENCE_GENOME_hg38 = "ref/hg38.fa"
